//******************************************************************************
//
// File:    StatTest03.java
// Package: ---
// Unit:    Class StatTest03
//
// This Java source file is copyright (C) 2017 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This Java source file is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html.
//
//******************************************************************************

import edu.rit.io.InStream;
import edu.rit.pj2.LongLoop;
import edu.rit.pj2.Task;
import edu.rit.pj2.vbl.LongVbl;
import edu.rit.util.Random;
import edu.rit.util.RandomSubset;
import java.io.File;
import java.io.FileInputStream;

/**
 * Class StatTest03 is a multicore parallel main program that performs a
 * statistical test on the LC3 cipher.
 * <P>
 * Usage: <TT>java pj2 StatTest03 [-v] <I>dfreq</I> <I>Nlen</I> <I>Plen</I>
 * <I>Slen</I> <I>reps</I> <I>seed</I></TT>
 * <P>
 * The program performs <I>reps</I> repetitions in parallel. For each
 * repetition, the program picks a random key; a random nonce of length
 * <I>Nlen</I>; and a random plaintext of length <I>Plen</I> using the digram
 * frequencies in the <I>dfreq</I> file. The final <I>Slen</I> letters of the
 * plaintext constitute the signature. The program computes the ciphertext from
 * the key, nonce, and plaintext. The program alters one letter in the nonce or
 * the ciphertext, decrypts the result, checks whether the decrypted signature
 * is the same as the original signature (a false positive authentication), and
 * repeats for every possible alteration. For each nonce and ciphertext
 * position, the program counts how many alterations resulted in a false
 * positive authentication and computes the probability of such.
 * <P>
 * The program prints the total number of false positive authentications and the
 * probability of such over all the nonce and ciphertext positions and over all
 * the repetitions. If the <TT>-v</TT> flag is specified, the program also
 * prints this data for each individual nonce and ciphertext position.
 *
 * @author  Alan Kaminsky
 * @version 24-Mar-2017
 */
public class StatTest03
	extends Task
	{

	// Command line arguments.
	boolean verbose;
	File dfreq;
	int Nlen;
	int Plen;
	int Slen;
	long reps;
	long seed;

	// Global variables.
	DigramHist generatorHist;

	// False positive counters for each nonce and ciphertext position.
	int Mlen;
	LongVbl[] fpCount;
	LongVbl totalFpCount;

	/**
	 * Main program.
	 */
	public void main
		(String[] args)
		throws Exception
		{
		// Parse command line arguments.
		int argi = 0;
		while (argi < args.length && args[argi].charAt(0) == '-')
			{
			switch (args[argi])
				{
				case "-v":
					verbose = true;
					break;
				default:
					usage();
				}
			++ argi;
			}
		if (args.length - argi != 6) usage();
		dfreq = new File (args[argi]);
		Nlen = Integer.parseInt (args[argi+1]);
		Plen = Integer.parseInt (args[argi+2]);
		Slen = Integer.parseInt (args[argi+3]);
		reps = Long.parseLong (args[argi+4]);
		seed = Long.parseLong (args[argi+5]);

		// Print provenance.
		System.out.printf ("$ java pj2 StatTest03");
		for (String arg : args) System.out.printf (" %s", arg);
		System.out.println();
		System.out.printf ("Nonce length = %d%n", Nlen);
		System.out.printf ("Plaintext length = %d%n", Plen);
		System.out.printf ("Signature length = %d%n", Slen);
		System.out.printf ("Repetitions = %d%n", reps);
		System.out.flush();

		// Read digram probabilities from file.
		InStream in = new InStream (new FileInputStream (dfreq));
		generatorHist = (DigramHist) in.readObject();
		in.close();

		// Set up storage for false positive counters.
		Mlen = Nlen + Plen;
		fpCount = new LongVbl [Mlen];
		for (int i = 0; i < Mlen; ++ i)
			fpCount[i] = new LongVbl.Sum();
		totalFpCount = new LongVbl.Sum();

		// Do repetitions in parallel.
		parallelFor (0, reps - 1) .schedule (dynamic) .exec (new LongLoop()
			{
			// Per-thread variables.
			int[] key;
			int[] origNonce;
			int[] origPlaintext;
			int[] origCiphertext;
			int[] badNonce;
			int[] badCiphertext;
			int[] badPlaintext;
			LC4 cipher;
			Random prng;
			RandomSubset rs;
			DigramPicker dp;
			LongVbl[] thrFpCount;
			LongVbl thrTotalFpCount;

			// Initialize per-thread variables.
			public void start()
				{
				// Set up key, nonce, plaintext, ciphertext.
				key = new int [36];
				origNonce = new int [Nlen];
				origPlaintext = new int [Plen];
				origCiphertext = new int [Plen];
				badNonce = new int [Nlen];
				badCiphertext = new int [Plen];
				badPlaintext = new int [Plen];

				// Set up cipher.
				cipher = new LC4();

				// Set up pseudorandom number generator.
				prng = new Random (seed + rank());
				rs = new RandomSubset (prng, 36, true);
				dp = new DigramPicker (prng, generatorHist);

				// Set up per-thread false positive counters.
				thrFpCount = new LongVbl [Mlen];
				for (int i = 0; i < Mlen; ++ i)
					thrFpCount[i] = threadLocal (fpCount[i]);
				thrTotalFpCount = threadLocal (totalFpCount);
				}

			// Perform one repetition.
			public void run (long rep)
				{
				// Generate random key.
				Util.genKey (key, rs);

				// Generate random original nonce.
				Util.genNonce (origNonce, prng);
				System.arraycopy (origNonce, 0, badNonce, 0, Nlen);

				// Generate random original plaintext.
				Util.genPlaintext (origPlaintext, dp);

				// Compute original ciphertext.
				cipher.encrypt (key, origNonce, origPlaintext, origCiphertext);
				System.arraycopy (origCiphertext, 0, badCiphertext, 0, Plen);

				// Consider each possible alteration to the nonce.
				for (int i = 0; i < Nlen; ++ i)
					{
					for (int j = 0; j < 35; ++ j)
						{
						badNonce[i] = (badNonce[i] + 1) % 36;
						checkFalsePositive (i);
						}
					badNonce[i] = origNonce[i];
					}

				// Consider each possible alteration to the ciphertext.
				for (int i = 0; i < Plen; ++ i)
					{
					for (int j = 0; j < 35; ++ j)
						{
						badCiphertext[i] = (badCiphertext[i] + 1) % 36;
						checkFalsePositive (Nlen + i);
						}
					badCiphertext[i] = origCiphertext[i];
					}
				}

			// Check for a false positive with an alteration at index a.
			private void checkFalsePositive (int a)
				{
				cipher.decrypt (key, badNonce, badCiphertext, badPlaintext);
				for (int i = Plen - Slen; i < Plen; ++ i)
					if (badPlaintext[i] != origPlaintext[i])
						return;
				++ thrFpCount[a].item;
				++ thrTotalFpCount.item;
				}
			});

		// Print false positive counts and probabilities.
		long denom = reps*35;
		System.out.printf ("\tFalse Positive%n");
		System.out.printf ("Posn\tCount\tProb%n");
		if (verbose)
			for (int i = 0; i < Mlen; ++ i)
				System.out.printf ("%d\t%d\t%.5g%n", i, fpCount[i].item,
					(double)(fpCount[i].item)/(double)denom);
		denom *= Mlen;
		System.out.printf ("Total\t%d\t%.5g%n", totalFpCount.item,
			(double)(totalFpCount.item)/(double)denom);
		}

	/**
	 * Print a usage message and exit.
	 */
	private static void usage()
		{
		System.err.println ("Usage: java pj2 StatTest03 [-v] <dfreq> <Nlen> <Plen> <Slen> <reps> <seed>");
		System.err.println ("-v = Print verbose output");
		System.err.println ("<dfreq> = File of digram frequencies");
		System.err.println ("<Nlen> = Nonce length");
		System.err.println ("<Plen> = Plaintext+signature length");
		System.err.println ("<Slen> = Signature length");
		System.err.println ("<reps> = Number of repetitions");
		System.err.println ("<seed> = Random seed");
		System.exit (1);
		}

	}
