//******************************************************************************
//
// File:    State.java
// Package: ---
// Unit:    Class State
//
// This Java source file is copyright (C) 2017 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This Java source file is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html.
//
//******************************************************************************

import edu.rit.pj2.vbl.LongVbl;
import edu.rit.util.BitSet64;

/**
 * Class State provides an encryption state object used in a
 * state recovery attack on the LC4 cipher.
 *
 * @author  Alan Kaminsky
 * @version 29-Mar-2017
 */
public class State
	{
	private static final int N = 6;
	private static final int NSQR = N*N;

	// True to print a trace during search.
	public static boolean trace = false;

	// Plaintext length, plaintext, ciphertext.
	private int M;
	private int[] plaintext;
	private int[] ciphertext;

	// LC4 state matrix. Unknown element = -1.
	private int[][] state = new int [N] [N];

	// LC4 marker.
	private int krow, kcol;

	// Index of message letter being verified.
	private int mi;

	// Row/column of plaintext/ciphertext letter.
	private int prow, pcol, crow, ccol;

	// Set of missing key letters.
	private BitSet64 missing = new BitSet64();

	// Number of conflicts.
	private LongVbl conflicts = new LongVbl();

	// Create a new LC4 state object with the given plaintext, the given
	// ciphertext, the given key, and the given number of missing key letters.
	public State
		(int[] plaintext,
		 int[] ciphertext,
		 int[] key,
		 int K)
		{
		M = plaintext.length;
		this.plaintext = (int[]) plaintext.clone();
		this.ciphertext = (int[]) ciphertext.clone();
		for (int i = 0; i < N; ++ i)
			for (int j = 0; j < N; ++ j)
				{
				int index = N*i + j;
				if (index < NSQR - K)
					{
					state[i][j] = key[index];
					}
				else
					{
					state[i][j] = -1;
					missing.add (key[index]);
					}
				}
		}

	// Create a new LC4 state object that is a copy of the given object.
	public State
		(State s)
		{
		this.M = s.M;
		this.plaintext = s.plaintext;
		this.ciphertext = s.ciphertext;
		for (int i = 0; i < N; ++ i)
			this.state[i] = (int[]) s.state[i].clone();
		this.krow = s.krow;
		this.kcol = s.kcol;
		this.mi = s.mi;
		this.prow = s.prow;
		this.pcol = s.pcol;
		this.crow = s.crow;
		this.ccol = s.ccol;
		this.missing.copy (s.missing);
		this.conflicts = s.conflicts;
		}

	// Returns this state as an array.
	public int[] getKey()
		{
		int[] key = new int [NSQR];
		for (int i = 0; i < N; ++ i)
			for (int j = 0; j < N; ++ j)
				key[N*i+j] = state[i][j];
		return key;
		}

	// Returns the number of conflicts found during a call to verify().
	public long conflictCount()
		{
		return conflicts.item;
		}

	// Verify state is correct for message index mi. Returns initial state
	// if verified, null if not verified.
	public State verify()
		{
		if (trace && mi == 0)
			{
			System.out.printf ("verify()%n");
			System.out.printf ("    ");
			printState();
			System.out.println();
			}

		// Base case.
		if (mi == M)
			{
			return recover();
			}

		// Recursive case.
		State ok = null;

		// Locate plaintext letter in state.
		int pval = plaintext[mi];
		prow = getRow (pval);
		pcol = getCol (pval);

		// If plaintext letter is not found, guess its location.
		if (prow == -1)
			{
			missing.remove (pval);
			for (prow = 0; prow < N && ok == null; ++ prow)
				for (pcol = 0; pcol < N && ok == null; ++ pcol)
					if (state[prow][pcol] == -1)
						ok = new State (this) .guessPlaintextLetter (pval);
			missing.add (pval);
			}

		// If plaintext letter is found, proceed.
		else
			{
			ok = new State (this) .guessPlaintextLetter (pval);
			}

		return ok;
		}

	// Guess plaintext letter pval at state row prow, column pcol.
	private State guessPlaintextLetter
		(int pval)
		{
		State ok = null;
		state[prow][pcol] = pval;

		// Get marker letter.
		int marker = state[krow][kcol];

		// If marker letter is unknown, guess missing key letters.
		if (marker == -1)
			{
			for (marker = 0; marker < NSQR && ok == null; ++ marker)
				if (missing.contains (marker))
					{
					missing.remove (marker);
					ok = new State (this) .guessMarkerLetter (marker);
					missing.add (marker);
					}
			}

		// If marker letter is known, proceed.
		else
			{
			ok = new State (this) .guessMarkerLetter (marker);
			}

		return ok;
		}

	// Guess marker letter.
	private State guessMarkerLetter
		(int marker)
		{
		State ok = null;
		state[krow][kcol] = marker;

		// Get ciphertext letter.
		crow = increaseRowIndex (prow, marker);
		ccol = increaseColIndex (pcol, marker);
		int cval = state[crow][ccol];

		// If ciphertext letter is unknown, guess missing key letters.
		if (cval == -1)
			{
			for (cval = 0; cval < NSQR && ok == null; ++ cval)
				if (missing.contains (cval))
					{
					missing.remove (cval);
					ok = new State (this) .guessCiphertextLetter (cval);
					missing.add (cval);
					}
			}

		// If ciphertext letter is known, proceed.
		else
			{
			ok = new State (this) .guessCiphertextLetter (cval);
			}

		return ok;
		}

	// Guess ciphertext letter cval at row crow, column ccol.
	private State guessCiphertextLetter
		(int cval)
		{
		state[crow][ccol] = cval;

		// If ciphertext letter in state is incorrect, conflict.
		if (cval != ciphertext[mi])
			{
			++ conflicts.item;
			return null;
			}

		// Advance to next state.
		else
			{
			shiftRow (prow);
			shiftCol (ccol);
			krow = increaseRowIndex (krow, cval);
			kcol = increaseColIndex (kcol, cval);

			if (trace)
				{
				System.out.printf ("%-3d ", mi);
				printState();
				System.out.printf ("%-3c%-3c%n",
					Util.intToLetter (plaintext[mi]),
					Util.intToLetter (ciphertext[mi]));
				}

			++ mi;
			return verify();
			}
		}

	// Recover the original state by working backwards through the
	// plaintext and ciphertext.
	private State recover()
		{
		if (trace)
			System.out.printf ("recover()%n");

		do
			{
			-- mi;
			int pval = plaintext[mi];
			prow = getRow (pval);
			pcol = getCol (pval);
			int cval = ciphertext[mi];
			crow = getRow (cval);
			ccol = getCol (cval);

			if (trace)
				{
				System.out.printf ("%-3d ", mi);
				printState();
				System.out.printf ("%-3c%-3c%n",
					Util.intToLetter (pval), Util.intToLetter (cval));
				}

			kcol = decreaseColIndex (kcol, cval);
			krow = decreaseRowIndex (krow, cval);
			shiftColUp (ccol);
			shiftRowLeft (prow);
			}
		while (mi > 0);

		if (trace)
			{
			System.out.printf ("    ");
			printState();
			System.out.println();
			}

		return this;
		}

	// Returns the state row in which the given integer appears, or -1 if
	// not found.
	private int getRow
		(int c)
		{
		for (int i = 0; i < N; ++ i)
			for (int j = 0; j < N; ++ j)
				if (state[i][j] == c)
					return i;
		return -1;
		}

	// Returns the state column in which the given integer appears, or -1 if
	// not found.
	private int getCol
		(int c)
		{
		for (int i = 0; i < N; ++ i)
			for (int j = 0; j < N; ++ j)
				if (state[i][j] == c)
					return j;
		return -1;
		}

	// Shift the given state row one place right.
	private void shiftRow
		(int r)
		{
		int t = state[r][N-1];
		for (int c = N - 1; c > 0; -- c)
			state[r][c] = state[r][c-1];
		state[r][0] = t;
		if (krow == r) kcol = incrIndex (kcol);
		if (prow == r) pcol = incrIndex (pcol);
		if (crow == r) ccol = incrIndex (ccol);
		}

	// Shift the given state row one place left.
	private void shiftRowLeft
		(int r)
		{
		int t = state[r][0];
		for (int c = 0; c < N - 1; ++ c)
			state[r][c] = state[r][c+1];
		state[r][N-1] = t;
		if (krow == r) kcol = decrIndex (kcol);
		if (prow == r) pcol = decrIndex (pcol);
		if (crow == r) ccol = decrIndex (ccol);
		}

	/**
	 * Shift the given state column one place down.
	 */
	private void shiftCol
		(int c)
		{
		int t = state[N-1][c];
		for (int r = N - 1; r > 0; -- r)
			state[r][c] = state[r-1][c];
		state[0][c] = t;
		if (kcol == c) krow = incrIndex (krow);
		if (pcol == c) prow = incrIndex (prow);
		if (ccol == c) crow = incrIndex (crow);
		}

	/**
	 * Shift the given state column one place up.
	 */
	private void shiftColUp
		(int c)
		{
		int t = state[0][c];
		for (int r = 0; r < N - 1; ++ r)
			state[r][c] = state[r+1][c];
		state[N-1][c] = t;
		if (kcol == c) krow = decrIndex (krow);
		if (pcol == c) prow = decrIndex (prow);
		if (ccol == c) crow = decrIndex (crow);
		}

	/**
	 * Returns row or column index x incremented.
	 */
	private static int incrIndex
		(int x)
		{
		return (x + 1)%N;
		}

	/**
	 * Returns row or column index x decremented.
	 */
	private static int decrIndex
		(int x)
		{
		return (x - 1 + N)%N;
		}

	/**
	 * Returns row index r increased as determined by value v.
	 */
	private static int increaseRowIndex
		(int r,
		 int v)
		{
		return (r + v/N)%N;
		}

	/**
	 * Returns row index r decreased as determined by value v.
	 */
	private static int decreaseRowIndex
		(int r,
		 int v)
		{
		return (r - v/N + N)%N;
		}

	/**
	 * Returns column index c increased as determined by value v.
	 */
	private static int increaseColIndex
		(int c,
		 int v)
		{
		return (c + v%N)%N;
		}

	/**
	 * Returns column index c decreased as determined by value v.
	 */
	private static int decreaseColIndex
		(int c,
		 int v)
		{
		return (c - v%N + N)%N;
		}

	/**
	 * Print the state.
	 */
	private void printState()
		{
		for (int i = 0; i < N; ++ i)
			{
			for (int j = 0; j < N; ++ j)
				System.out.printf ("%c", Util.intToLetter (state[i][j]));
			System.out.printf (" ");
			}
		System.out.printf (" %-3d%-3d", krow, kcol);
		}
	}
