//******************************************************************************
//
// File:    StateRecovery01.java
// Package: ---
// Unit:    Class StateRecovery01
//
// This Java source file is copyright (C) 2017 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This Java source file is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html.
//
//******************************************************************************

import edu.rit.io.InStream;
import edu.rit.util.Random;
import edu.rit.util.RandomSubset;
import java.io.File;
import java.io.FileInputStream;
import java.util.Arrays;

/**
 * Class StateRecovery01 implements a state recovery attack on the LC4 cipher.
 * <P>
 * Usage: <TT>java StateRecovery01 [-v] <I>dfreq</I> <I>K</I> <I>N</I> <I>P</I>
 * <I>seed</I></TT>
 * <P>
 * The program picks a random key; a random nonce of length <I>Nlen</I>; and a
 * random plaintext of length <I>Plen</I> using the digram frequencies in the
 * <I>dfreq</I> file. The program computes the ciphertext. The program "forgets"
 * the final <I>K</I> letters of the key. The program does a branch-and-bound
 * search to recover the missing key letters and reports the number of conflicts
 * encountered during the search. If the <TT>-v</TT> flag is specified, the
 * program prints the state of the search at each step.
 *
 * @author  Alan Kaminsky
 * @version 29-Mar-2017
 */
public class StateRecovery01
	{
	private static final int N = 6;
	private static final int NSQR = N*N;

	// Command line arguments.
	static boolean vflag;
	static File dfreq;
	static int K;
	static int E;
	static int P;
	static long seed;

	// Global variables.
	static int[] origKey;
	static int M;
	static int[] plaintext;
	static int[] ciphertext;

	/**
	 * Main program.
	 */
	public static void main
		(String[] args)
		throws Exception
		{
		// Parse command line arguments.
		int argi = 0;
		while (argi < args.length && args[argi].startsWith ("-"))
			{
			if (args[argi].equals ("-v"))
				{
				vflag = true;
				++ argi;
				}
			else
				usage();
			}
		if (args.length - argi != 5) usage();
		dfreq = new File (args[argi+0]);
		K = Integer.parseInt (args[argi+1]);
		E = Integer.parseInt (args[argi+2]);
		P = Integer.parseInt (args[argi+3]);
		seed = Long.parseLong (args[argi+4]);

		// Print provenance.
		System.out.printf ("$ java StateRecovery01");
		for (String arg : args) System.out.printf (" %s", arg);
		System.out.println();

		// Read digram probabilities from file.
		InStream in = new InStream (new FileInputStream (dfreq));
		DigramHist generatorHist = (DigramHist) in.readObject();
		in.close();

		// Generate random key.
		Random prng = new Random (seed);
		RandomSubset rs = new RandomSubset (prng, NSQR, true);
		origKey = new int [NSQR];
		Util.genKey (origKey, rs);

		// Generate random plaintext and nonce.
		M = E + P;
		plaintext = new int [M];
		DigramPicker dp = new DigramPicker (prng, generatorHist);
		Util.genPlaintext (plaintext, dp);
		for (int i = 0; i < E; ++ i)
			plaintext[i] = prng.nextInt (NSQR);

		// Compute ciphertext.
		LC4 cipher = new LC4();
		cipher.trace = vflag;
		ciphertext = new int [M];
		cipher.encrypt (origKey, new int [0], plaintext, ciphertext);

		// Print inputs.
		printLetters (origKey, "Key");
		printLetters (plaintext, "Plaintext ");
		printLetters (ciphertext, "Ciphertext");
		System.out.printf ("%d missing key letters%n", K);

		// Find state.
		State origState = new State (plaintext, ciphertext, origKey, K);
		origState.trace = vflag;
		State recovState = origState.verify();
		if (recovState == null)
			System.out.printf ("Recovery failed!%n");
		else
			{
			int[] recovKey = recovState.getKey();
			printLetters (recovKey, "Recovered key");
			if (Arrays.equals (origKey, recovKey))
				System.out.printf ("Recovered key matches!%n");
			else
				System.out.printf ("Recovered key doesn't match!%n");
			}
		System.out.printf ("%d conflicts%n", recovState.conflictCount());
		}

	/**
	 * Print the given integer array as a string of letters.
	 */
	private static void printLetters
		(int[] s,
		 String label)
		{
		System.out.printf ("%s = %s%n", label, Util.intsToLetters (s));
		}

	/**
	 * Print a usage message and exit.
	 */
	private static void usage()
		{
		System.err.println ("Usage: java StateRecovery01 [-v] <dfreq> <K> <N> <P> <seed>");
		System.err.println ("-v = Verbose printout");
		System.err.println ("<dfreq> = File of digram frequencies");
		System.err.println ("<K> = Number of missing key letters");
		System.err.println ("<N> = Nonce length");
		System.err.println ("<P> = Plaintext length");
		System.err.println ("<seed> = Random seed");
		System.exit (1);
		}

	}
